"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.getStringParameters = getStringParameters;
exports.getOptions = getOptions;
exports.getTypeDeclarationInBlock = getTypeDeclarationInBlock;
exports.getRegisterArguments = getRegisterArguments;
exports.getBlockParent = getBlockParent;
exports.getTypeParameter = getTypeParameter;
exports.throwMaybeAstError = exports.throwUnexpectedError = exports.Errors = exports.ErrorBase = void 0;

var _core = require("@babel/core");

var _babelPluginMacros = require("babel-plugin-macros");

var _commonTags = require("common-tags");

var _fastCopy = _interopRequireDefault(require("fast-copy"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 *  This file contains compile time helpers to validate that the macro is being called appropriately.
 */
// This is used in order to reduce duplication in the compile error tests
// If you update a message in here, the corresponding compile error test will pass automatically.
// No updating needed.
const ErrorBase = {
  ValidatorNoTypeParameter: `Failed to find type parameter. createValidator should be called with a type parameter. Example: "createValidator<TypeName>()"`,
  NotCalledAsFunction: `Macro should be called as function but was called as a`,
  MoreThanOneTypeParameter: `Macro should be called with 1 type parameter but was called with`,
  RegisterInvalidCall: (0, _commonTags.stripIndent)`
                                "register" should be called right below the type
                                declaration it is registering:
                                \`\`\`
                                interface Foo {
                                  bar: string
                                }
                                // no return value
                                register<Foo>()
                                \`\`\`
                                `,
  UnregisteredType: `Tried to generate a validator for an unregistered type with name`,
  // TODO: These will need to be updated once register accepts an options object
  RegisterInvalidNumberParams: `register should be called with 1 argument, but it was called with`,
  RegisterParam1NotStringLiteral: `register's first (and only) parameter should be a string literal, which is the name of the type to register, but it was a`,
  TypeDoesNotAcceptGenericParameters: `types don't accept generic parameters`,
  TooManyTypeParameters: `even though it only accepts`,
  NotEnoughTypeParameters: `type parameters even though it requires at least`,
  InvalidTypeParameterReference: `tried to reference the default type parameter in position:`
};
exports.ErrorBase = ErrorBase;

const removePeriod = str => str.slice(-1) === "." ? str.slice(0, -1) : str;

const addNewline = str => str.slice(-1) === "\n" ? str : str + "\n";

const Errors = {
  // Definitely user errors
  NoTypeParameters: () => ErrorBase.ValidatorNoTypeParameter,
  NotCalledAsFunction: callType => `${ErrorBase.NotCalledAsFunction} ${callType}`,
  MoreThanOneTypeParameter: typeParams => `${ErrorBase.MoreThanOneTypeParameter} ${typeParams}`,
  InvalidRegisterCall: () => ErrorBase.RegisterInvalidCall,
  // Errors that indicate possible bugs in the macro itself
  InvalidNumberOfRegisterParams: params => `${ErrorBase.RegisterInvalidNumberParams} ${params}`,
  RegisterParam1NotStringLiteral: paramNode => `${ErrorBase.RegisterParam1NotStringLiteral} ${paramNode.type}`,
  UnregisteredType: typeName => `${ErrorBase.UnregisteredType}: ${typeName}`,
  TypeDoesNotAcceptGenericParameters: (typeName, nodeType) => `Tried to instantiate "${typeName}" with generic parameters even though ${nodeType} ${ErrorBase.TypeDoesNotAcceptGenericParameters}`,
  TooManyTypeParameters: (typeName, provided, actual) => `Tried to instantiate ${typeName} with ${provided} type parameters ${ErrorBase.TooManyTypeParameters} ${actual}`,
  NotEnoughTypeParameters: (typeName, provided, actual) => `Tried to instantiate ${typeName} with ${provided} ${ErrorBase.NotEnoughTypeParameters} ${actual}`,
  InvalidTypeParameterReference: (paramPosition, referencedPosition) => (0, _commonTags.oneLine)`The default type parameter in position: ${paramPosition}
    ${ErrorBase.InvalidTypeParameterReference} ${referencedPosition}, which has not yet been instantiated`,
  UnexpectedError: reason => {
    return (0, _commonTags.oneLine)`Unexpected error because ${removePeriod(reason)}. Please report this to the developer.`;
  },
  // TODO: Can remove optional parameter (probably)
  MaybeAstError: (reason, optional) => {
    return (0, _commonTags.oneLine)`Invalid AST node because ${removePeriod(reason)}. Most likely, your Typescript is invalid.
    ${optional !== undefined ? addNewline(optional) : ""}But if it isn't, please report this to the developer.`;
  }
};
exports.Errors = Errors;

const throwUnexpectedError = msg => {
  throw new _babelPluginMacros.MacroError(Errors.UnexpectedError(msg));
};

exports.throwUnexpectedError = throwUnexpectedError;

const throwMaybeAstError = (msg, optional) => {
  throw new _babelPluginMacros.MacroError(Errors.MaybeAstError(msg, optional));
};

exports.throwMaybeAstError = throwMaybeAstError;

function assertCallExpr(expr) {
  if (!expr.isCallExpression()) throw new _babelPluginMacros.MacroError(Errors.NotCalledAsFunction(expr.type));
}

function getStringParameters(macroPath, functionName) {
  const callExpr = macroPath.parentPath;
  assertCallExpr(callExpr);
  const args = callExpr.get("arguments");
  assertArray(args);
  const strings = [];

  for (const arg of args) {
    const {
      confident,
      value
    } = arg.evaluate();

    if (!confident || typeof value !== "string") {
      throwMaybeAstError(`the arguments to ${functionName} were not all string literals`);
    }

    strings.push(value);
  }

  return strings;
}

// Keep this in sync with dist/typecheck.macro.d.ts
const keys = {
  c: "circularRefs",
  e: "expectedValueAsIR",
  f: "allowForeignKeys"
};
const defaultOpts = {
  [keys.c]: true,
  [keys.f]: true,
  [keys.e]: false
};
Object.freeze(defaultOpts);

function getOptions(validatorType, macroPath, functionName) {
  const callExpr = macroPath.parentPath;
  assertCallExpr(callExpr);
  const args = callExpr.get("arguments");
  assertArray(args);

  if (args.length === 0) {
    return defaultOpts;
  }

  const notOptionsObjectMessage = `${functionName}'s sole argument is an options object that is statically known at COMPILE TIME.`;

  if (args.length > 1) {
    throw new _babelPluginMacros.MacroError(notOptionsObjectMessage);
  } else {
    const arg = args[0];
    const {
      confident,
      value
    } = arg.evaluate();

    if (!confident || typeof value !== "object" || value === null) {
      throw new _babelPluginMacros.MacroError(notOptionsObjectMessage);
    }

    const opts = (0, _fastCopy.default)(defaultOpts);
    applyOption(opts, value, keys.c, ["boolean", "undefined"]);

    if (validatorType === "detailed") {
      applyOption(opts, value, keys.e, ["boolean", "undefined"]);
    }

    applyOption(opts, value, keys.f, ["boolean", "undefined"]);
    return opts;
  }
}

function applyOption(opts, value, key, allowedTypes) {
  const type = typeof value[key];

  for (const allowedType of allowedTypes) {
    if (type === allowedType) {
      opts[key] = type === "undefined" ? defaultOpts[key] : value[key];
      return;
    }
  }

  throw new _babelPluginMacros.MacroError(`type of options.${key} should be ${(0, _commonTags.commaListsOr)`${allowedTypes}`}, but it was: "${type}"`);
}

function assertSingular(expr) {
  if (expr === null || expr === undefined || Array.isArray(expr)) {
    throw new _babelPluginMacros.MacroError(Errors.UnexpectedError(`expected expr to be single NodePath but it was ${expr}`));
  }
}

function assertArray(expr) {
  if (expr === null || expr === undefined || !Array.isArray(expr)) {
    throwUnexpectedError(`Expected expression to be array, but it was ${JSON.stringify(expr)}`);
  }
}

function getTypeDeclarationInBlock(typeName, stmts, idxInBlock) {
  for (let i = 0; i < stmts.length; ++i) {
    if (i === idxInBlock) continue;
    let stmt = stmts[i]; // If the statement is an export (named or default) use the declared statement instead

    if (_core.types.isExportNamedDeclaration(stmt) || _core.types.isExportDefaultDeclaration(stmt)) {
      stmt = stmt.declaration;
    }

    if (_core.types.isTSInterfaceDeclaration(stmt) || _core.types.isTSTypeAliasDeclaration(stmt)) {
      const declarationName = stmt.id.name;

      if (declarationName === typeName) {
        return stmt;
      }
    }
  }

  return null;
}

function getRegisterArguments(macroPath) {
  const callExpr = macroPath.parentPath;
  assertCallExpr(callExpr);
  const args = callExpr.node.arguments;

  if (args.length != 1) {
    throw new _babelPluginMacros.MacroError(Errors.InvalidNumberOfRegisterParams(args.length));
  }

  const typeNamePath = callExpr.get("arguments.0");
  assertSingular(typeNamePath);

  if (!typeNamePath.isStringLiteral()) {
    throw new _babelPluginMacros.MacroError(Errors.RegisterParam1NotStringLiteral(typeNamePath.node));
  }

  const {
    confident,
    value: typeName
  } = typeNamePath.evaluate();

  if (confident === false || typeof typeName !== "string") {
    throwUnexpectedError(`evaluation result had type ${typeof typeName} and confidence ${confident}`);
  }

  return typeName;
}

function getBlockParent(macroPath) {
  const callExpr = macroPath.parentPath;
  assertCallExpr(callExpr);
  const exprStmt = callExpr.parentPath;
  if (!exprStmt.isExpressionStatement()) throw new _babelPluginMacros.MacroError(Errors.InvalidRegisterCall());
  const {
    node
  } = exprStmt.parentPath;

  if (!_core.types.isProgram(node) && !_core.types.isBlock(node)) {
    throw new _babelPluginMacros.MacroError(Errors.InvalidRegisterCall());
  } else {
    return node.body;
  }
}

function getTypeParameter(macroPath) {
  const callExpr = macroPath.parentPath;
  assertCallExpr(callExpr);
  const typeParametersPath = callExpr.get("typeParameters");

  if (typeParametersPath && !Array.isArray(typeParametersPath) && typeParametersPath.node) {
    const {
      node
    } = typeParametersPath;

    if (_core.types.isTSTypeParameterInstantiation(node)) {
      const params = node.params.length;
      if (params != 1) throw new _babelPluginMacros.MacroError(Errors.MoreThanOneTypeParameter(params));
      const typeParameterPath = typeParametersPath.get("params.0");

      if (!Array.isArray(typeParameterPath) && // @ts-ignore: https://github.com/babel/babel/issues/11535
      typeParameterPath.isTSType()) {
        // @ts-ignore: https://github.com/babel/babel/issues/11535
        return typeParameterPath;
      }

      throwUnexpectedError(`typeParameter was ${Array.isArray(typeParameterPath) ? "an array" : "not a TSType"}`);
    }

    throwUnexpectedError(`typeParameters node was ${node.type} instead of TSTypeParameterInstantiation`);
  } else {
    throw new _babelPluginMacros.MacroError(Errors.NoTypeParameters());
  }
}